#include "../Framework2/OpenGL/OpenGLApp.h"
#include "../Framework2/Util/OpenGLModel.h"
#include "../Framework2/Util/CollisionSet.h"
#include "../Framework2/Util/ParticleSystem.h"
#include "../Framework2/Util/DemoTrack.h"
#include "../Framework2/Math/PerlinNoise.h"

#define SIZE_Y  64
#define SIZE_XZ (2 * SIZE_Y)

#define WORLD_SIZE 1024.0f

class MainApp : public OpenGLApp {
public:
	void resetCamera();
	void processMovement(const vec3 &newPosition);

//	void setKey(unsigned int key, bool value);
	void initMenu();

	bool init();
	bool exit();

	bool load();
	bool unload();

	void drawLight(const vec3 &lightPos, float size, const vec3 &color);
	void drawFire(char *vertices);
	void drawSmoke(char *vertices);
	void drawParticles();
	bool drawFrame();
protected:
	CollisionSet coll;
	vec3 lightPos;

	ShaderID lighting, smokeLighting, ufoLighting;
	OpenGLModel *room, *ufo;

	ParticleSystem *fire, *smoke;

	DemoTrack demoTrack;

	TextureID base[3], bump[3], lightVolume;
	TextureID particle, ufoBase;

	bool drawCoronas;
};

void MainApp::resetCamera(){
	position = vec3(430, -160, 0);
	wx = 0;
	wy = 2.43f;
	wz = 0;
}

void MainApp::processMovement(const vec3 &newPosition){
	vec3 pos = newPosition;

	coll.pushSphere(pos, 42);

	position = pos;
}
/*
void MainApp::setKey(unsigned int key, bool value){
	App::setKey(key, value);

	if (value){
		if (key == KEY_INSERT){
			demoTrack.add(position);
		} else if (key == 'S'){
			demoTrack.writeToFile("ufoPath.pth");
		} else if (key == 'O'){
			demoTrack.loadFromFile("ufoPath.pth");
		}
	}
}
*/
void MainApp::initMenu(){
	Menu *menu = menuSystem->getMainMenu();
	menu->addMenuItem("Draw coronas: ", &drawCoronas, INPUT_BOOL);

	App::initMenu();
}

bool MainApp::init(){
	initPerlin();

	unsigned int i;

	ufo = new OpenGLModel();
	ufo->loadFromFile("../Models/UFO2.hmdl");
	for (i = 0; i < ufo->getBatchCount(); i++){
		Batch *batch = ufo->getBatch(i);
		batch->addNormals();
	}

	room = new OpenGLModel();
	room->loadFromFile("../Models/Room/Map.hmdl");

	for (i = 0; i < room->getBatchCount(); i++){
		Batch *batch = room->getBatch(i);

		char *srcVert = batch->getVertices();
		unsigned short *srcInd = (unsigned short *) batch->getIndices();
		unsigned int size = batch->getVertexSize(); 

		for (unsigned int j = 0; j < batch->getIndexCount(); j += 3){
			coll.addTriangle(*(vec3 *) (srcVert + srcInd[j] * size), *(vec3 *) (srcVert + srcInd[j + 1] * size), *(vec3 *) (srcVert + srcInd[j + 2] * size));
		}
	}
	coll.build();

	lightPos = vec3(0, 150, -512);


	fire = new ParticleSystem();
	fire->setSpawnRate(50);
	fire->setSpeed(250, 10);
	fire->setLife(1.0f, 0);
	fire->setDirectionalForce(vec3(0, 100, 0));
	fire->setFrictionFactor(1.0f);
	fire->setPosition(vec3(0, 0, -512));
	fire->setSize(50, 25);

	for (i = 0; i < 6; i++){
		fire->setColor(i, vec4(0.10f * i, 0.02f * i, 0, 0));
		fire->setColor(6 + i, vec4(0.10f * 6, 0.10f * i + 0.12f, 0.04f * i, 0));
	}

	smoke = new ParticleSystem();
	smoke->setSpawnRate(20);
	smoke->setSpeed(450, 0);
	smoke->setLife(3.0f, 0);
	smoke->setDirectionalForce(vec3(0, -100, 0));
	smoke->setFrictionFactor(0.98f);
	smoke->setPosition(vec3(0, -400, 768));
	smoke->setSize(180, 0);

	for (i = 0; i < 12; i++){
		smoke->setColor(i, vec4(0.5f, 0.5f, 0.5f, i / 11.0f));
	}

	demoTrack.loadFromFile("ufoPath.pth");
	//demoTrack.setNodeSize(3);

	drawCoronas = false;

	return true;
}

bool MainApp::exit(){
	delete fire;
	delete smoke;

	delete ufo;
	delete room;
	return true;
}

bool MainApp::load(){
	if (!GL_ARB_shader_objects_supported || !GL_ARB_vertex_shader_supported || !GL_ARB_fragment_shader_supported || !GL_ARB_shading_language_100_supported){
		addToLog("No GLSL support (GL_ARB_shader_objects, GL_ARB_vertex_shader, GL_ARB_fragment_shader, GL_ARB_shading_language_100)\n");
		return false;
	}

	if ((lighting      = renderer->addShader("lighting.shd")) == SHADER_NONE) return false;
	if ((smokeLighting = renderer->addShader("smoke.shd")) == SHADER_NONE) return false;
	if ((ufoLighting   = renderer->addShader("ufo.shd")) == SHADER_NONE) return false;

	setDefaultFont("../Textures/Fonts/Future.font", "../Textures/Fonts/Future.dds");

	if ((base[0] = renderer->addTexture("../Textures/floor.dds")) == TEXTURE_NONE) return false;
	if ((bump[0] = renderer->addTexture("../Textures/floorBump.png", TEX_NORMALMAP)) == TEXTURE_NONE) return false;

	if ((base[1] = renderer->addTexture("../Textures/synbr05.dds")) == TEXTURE_NONE) return false;
	if ((bump[1] = renderer->addTexture("../Textures/synbr01Bump.png", TEX_NORMALMAP)) == TEXTURE_NONE) return false;

	if ((base[2] = renderer->addTexture("../Textures/stone34.dds")) == TEXTURE_NONE) return false;
	if ((bump[2] = renderer->addTexture("../Textures/stone34Bump.png", TEX_NORMALMAP)) == TEXTURE_NONE) return false;

	if ((particle = renderer->addTexture("../Textures/Particle.png")) == TEXTURE_NONE) return false;

	if ((ufoBase = renderer->addTexture("../Textures/ufo.dds")) == TEXTURE_NONE) return false;


	Image3D *img = new Image3D();
	if (!img->loadFromFile("lightMap.dds")){
		const char *text = "Precomputing the light volume map.\nThis will take some time.\n";
#ifdef _WIN32
		MessageBox(hwnd, text, "Precomputing", MB_OK | MB_ICONINFORMATION);
#else
		printf(text);
#endif

		unsigned char *lightMap = new unsigned char[SIZE_XZ * SIZE_XZ * SIZE_Y];
		unsigned char *dest = lightMap;

#define N_SAMPLES 100
		vec3 samples[N_SAMPLES];
		for (int a = 0; a < N_SAMPLES; a++){
			vec3 sample;
			do {
				sample = vec3(float(rand()), float(rand()), float(rand())) * (2.0f / RAND_MAX) - 1.0f;
			} while (dot(sample, sample) < 1.0f);

			samples[a] = sample * vec3(30, 60, 30);
		}

		for (int z = 0; z < SIZE_XZ; z++){
			for (int y = 0; y < SIZE_Y; y++){
				for (int x = 0; x < SIZE_XZ; x++){
					vec3 pos = vec3(x - 0.5f * (SIZE_XZ - 1), y - 0.5f * (SIZE_Y - 1), z - 0.5f * (SIZE_XZ - 1)) * WORLD_SIZE / SIZE_Y;

					coll.pushSphere(pos, WORLD_SIZE / SIZE_Y);

					float sum = 0;
					for (int i = 0; i < N_SAMPLES; i++){
						//vec3 rnd0 = (vec3(float(rand()), float(rand()), float(rand())) * (1.0f / RAND_MAX) - 0.5f) * (0.97f * WORLD_SIZE / SIZE_Y);
						//vec3 rnd1 = vec3(float(rand()), 2 * float(rand()), float(rand())) * (60.0f / RAND_MAX) - 30.0f;

						if (!coll.intersect(pos/* + rnd0*/, lightPos + samples[i], NULL, false)) sum++;
					}
					*dest++ = (unsigned char) (255.0f * sum / N_SAMPLES);
				}
			}
		}
		img->loadFromMemory(lightMap, SIZE_XZ, SIZE_Y, SIZE_XZ, FORMAT_I8, true, false);
		img->saveToFile("lightMap.dds");
	}
/*

#define getP(a,b,c) src[(min(max(a, 0), SIZE_XZ - 1) * SIZE_Y + min(max(b, 0), SIZE_Y - 1)) * SIZE_XZ + min(max(c, 0), SIZE_XZ - 1)]
	
	unsigned char *lMap = new unsigned char[SIZE_XZ * SIZE_XZ * SIZE_Y];
	unsigned char *dest = lMap;
	unsigned char *src = img->getImage();

	for (int z = 0; z < SIZE_XZ; z++){
		for (int y = 0; y < SIZE_Y; y++){
			for (int x = 0; x < SIZE_XZ; x++){

				int sum = 0;
				for (int c = -1; c <= 1; c++){
					for (int b = -1; b <= 1; b++){
						for (int a = -1; a <= 1; a++){
							sum += getP(z + c, y + b, x + a);
						}
					}
				}
				*dest++ = sum / 27;
			}
		}
	}
	img->loadFromMemory(lMap, SIZE_XZ, SIZE_Y, SIZE_XZ, FORMAT_I8, true, false);
*/

/*
	img->createMipMaps();
	Image3D *img2 = new Image3D();
	img2->loadFromMemory(img->getImage(1), img->getWidth(1), img->getHeight(1), img->getDepth(1), img->getFormat(), false, false);
*/

	if ((lightVolume = renderer->addTexture3D(img, TEX_NOMIPMAPPING | TEX_CLAMP/* | TEX_NOLINEARFILTER*/)) == TEXTURE_NONE) return false;

	room->uploadToVertexBuffer();
	ufo->uploadToVertexBuffer();

	return true;
}

bool MainApp::unload(){
	ufo->freeVertexBuffer();
	room->freeVertexBuffer();
	return true;
}

void MainApp::drawLight(const vec3 &lightPos, float size, const vec3 &color){
	renderer->setBlending(GL_ONE, GL_ONE);
	renderer->setTextures(particle);
	renderer->setMask(COLOR);
	renderer->apply();

	glColor3fv(color);

	vec3 dx(modelView.elem[0][0], modelView.elem[0][1], modelView.elem[0][2]);
	vec3 dy(modelView.elem[1][0], modelView.elem[1][1], modelView.elem[1][2]);
	glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex3fv(lightPos - size * dx + size * dy);

		glTexCoord2f(1,0);
		glVertex3fv(lightPos + size * dx + size * dy);

		glTexCoord2f(1,1);
		glVertex3fv(lightPos + size * dx - size * dy);

		glTexCoord2f(0,1);
		glVertex3fv(lightPos - size * dx - size * dy);
	glEnd();
}

void MainApp::drawFire(char *vertices){
	renderer->setTextures(particle);
	renderer->setBlending(ONE, ONE);
	renderer->setMask(COLOR);
	renderer->apply();

	glVertexPointer  (3, GL_FLOAT, sizeof(vec3) + sizeof(vec2) + sizeof(vec4), vertices);
	glTexCoordPointer(2, GL_FLOAT, sizeof(vec3) + sizeof(vec2) + sizeof(vec4), vertices + sizeof(vec3));
	glColorPointer   (4, GL_FLOAT, sizeof(vec3) + sizeof(vec2) + sizeof(vec4), vertices + sizeof(vec3) + sizeof(vec2));

	glDrawArrays(GL_QUADS, 0, fire->getParticleCount() * 4);
}

void MainApp::drawSmoke(char *vertices){
	renderer->setTexture("Base", particle);
	renderer->setTexture("LightVolume", lightVolume);
	renderer->setBlending(SRC_ALPHA, ONE_MINUS_SRC_ALPHA);
	renderer->setMask(COLOR);
	renderer->setShader(smokeLighting);
	renderer->apply();

	renderer->changeShaderConstant3f("scale", vec3(0.5f / WORLD_SIZE, 1.0f / WORLD_SIZE, 0.5f / WORLD_SIZE));

	glVertexPointer  (3, GL_FLOAT, sizeof(vec3) + sizeof(vec2) + sizeof(vec4), vertices);
	glTexCoordPointer(2, GL_FLOAT, sizeof(vec3) + sizeof(vec2) + sizeof(vec4), vertices + sizeof(vec3));
	glColorPointer   (4, GL_FLOAT, sizeof(vec3) + sizeof(vec2) + sizeof(vec4), vertices + sizeof(vec3) + sizeof(vec2));

	glDrawArrays(GL_QUADS, 0, smoke->getParticleCount() * 4);
}

void MainApp::drawParticles(){
	vec3 dx(modelView.elem[0][0], modelView.elem[0][1], modelView.elem[0][2]);
	vec3 dy(modelView.elem[1][0], modelView.elem[1][1], modelView.elem[1][2]);
	vec3 dz(modelView.elem[2][0], modelView.elem[2][1], modelView.elem[2][2]);

	fire->update(time);
	smoke->update(time);
	smoke->depthSort(position, dz);

	glClientActiveTextureARB(GL_TEXTURE0_ARB);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);

	if (position.z > 0){
		drawFire(fire->getVertexArray(dx, dy, true));
		drawSmoke(smoke->getVertexArray(dx, dy, true));
	} else {
		drawSmoke(smoke->getVertexArray(dx, dy, true));
		drawFire(fire->getVertexArray(dx, dy, true));
	}

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);

}

bool MainApp::drawFrame(){
	renderer->changeMask(ALL);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	projection = projectionMatrixX(1.5f, float(height) / float(width), 0.1f, 6000);
	glLoadMatrixf(transpose(projection));

	modelView = rotateZXY(-wx, -wy, -wz);
	modelView.translate(-position);

    glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(transpose(modelView));

	glEnable(GL_CULL_FACE);

	for (unsigned int i = 0; i < 3; i++){
		renderer->setShader(lighting);
		renderer->setTexture("Base", base[i]);
		renderer->setTexture("Bump", bump[i]);
		renderer->setTexture("LightVolume", lightVolume);
		renderer->apply();

		renderer->changeShaderConstant3f("lightPos", lightPos);
		renderer->changeShaderConstant3f("camPos", position);
		renderer->changeShaderConstant3f("scale", vec3(0.5f / WORLD_SIZE, 1.0f / WORLD_SIZE, 0.5f / WORLD_SIZE));

		((OpenGLBatch *) room->getBatch(i))->draw();
	}

	renderer->setShader(ufoLighting);
	renderer->setTexture("Base", ufoBase);
	renderer->setTexture("LightVolume", lightVolume);
	renderer->apply();

	vec3 ufoPos;
	//if (demoTrack.getNodeCount()){
		demoTrack.getAt(1.0f * time, (float *) &ufoPos, 0, 3);
	//}

	mat4 mat = translate(ufoPos) * rotateZXY(noise1(time), 30 * noise1(-0.3f * time), noise1(time + 13.41f));

	mat3 nMat(
		mat.elem[0][0], mat.elem[0][1], mat.elem[0][2], 
		mat.elem[1][0], mat.elem[1][1], mat.elem[1][2], 
		mat.elem[2][0], mat.elem[2][1], mat.elem[2][2]
	);

	renderer->changeShaderConstant3x3f("normMat", nMat);

	renderer->changeShaderConstant4x4f("modelMat", mat);
	renderer->changeShaderConstant3f("lightPos", lightPos);
	renderer->changeShaderConstant3f("camPos", position);
	renderer->changeShaderConstant3f("scale", vec3(0.5f / WORLD_SIZE, 1.0f / WORLD_SIZE, 0.5f / WORLD_SIZE));

	ufo->draw();
	
	drawParticles();

	if (drawCoronas){
		static float strength = 1.0f;
		if (coll.intersect(position, lightPos)){
			strength *= powf(0.01f, frameTime);
		} else {
			strength += 5 * frameTime;
			if (strength > 1.0f) strength = 1.0f;
		}
		if (strength > 0.01f){
			renderer->setDepthFunc(DEPTH_NONE);
			drawLight(lightPos, 480, 0.6f * strength * vec3(0.8f, 0.6f, 0));
		}
	}
	return true;
}

App *app = new MainApp();
